try:
    from collections import Mapping
except ImportError:
    from collections.abc import Mapping
from typing import List, Optional, BinaryIO, Dict

KEY_METHOD = 'http_method'
KEY_CONTENT_TYPE = 'content_type'
KEY_HEADERS = 'headers'
KEY_HTTP_HEADERS = 'http_headers'
KEY_CONTENTS = 'contents'
KEY_METADATA = 'metadata'
KEY_BODY = 'body'
KEY_FILENAME = 'file_name'
KEY_FORMNAME = 'form_name'


class Content(Mapping):

    def __init__(
            self,
            content_type: Optional[str],
            metadata: Optional[Dict],
            body: Optional[BinaryIO],
            file_name: Optional[str] = None,
            form_name: Optional[str] = None):
        self.__dict = {
            KEY_CONTENT_TYPE: content_type,
            KEY_METADATA: metadata,
            KEY_BODY: body,
            KEY_FILENAME: file_name,
            KEY_FORMNAME: form_name
        }

    def __getitem__(self, key):
        return self.__dict[key]

    def __iter__(self):
        return iter(self.__dict)

    def __len__(self):
        return len(self.__dict)

    def __contains__(self, item):
        return item in self.__dict

    def __str__(self):
        return str(self.__dict)

    def __eq__(self, o):
        return self.__dict.__eq__(o)

    def __ne__(self, o):
        return self.__dict.__ne__(o)

    def keys(self):
        return self.__dict.keys()

    def items(self):
        return self.__dict.items()

    def values(self):
        return self.__dict.values()

    def get(self, key, default=None):
        return self.__dict.get(key, default)

    def read(self, size: int = -1) -> bytes:
        body = self.__dict.get(KEY_BODY, None)
        if not body:
            raise ValueError('no body')
        return body.read(size)  # type: ignore

    def close(self):
        body = self.__dict.get(KEY_BODY, None)
        if body:
            body.close()


class Request(Mapping):

    def __init__(self, _method: str, _content_type: str, _headers: list = None, _contents: List[Content] = None):
        headers = _headers
        if not headers:
            headers = []
        contents = _contents
        if not contents:
            contents = []

        http_headers = HttpHeaders(headers)
        self.__dict = {
            KEY_METHOD: _method,
            KEY_CONTENT_TYPE: _content_type,
            KEY_HEADERS: headers,
            KEY_HTTP_HEADERS: http_headers,
            KEY_CONTENTS: contents
        }

    def append_content(self, content: Content) -> None:
        contents = self.__dict.get(KEY_CONTENTS, [])
        contents.append(content)  # type: ignore
        self.__dict[KEY_CONTENTS] = contents

    ###############################
    # methods for Mapping
    ###############################

    def __getitem__(self, key):
        if key == 'metadata':
            contents = self.__dict.get(KEY_CONTENTS, [])
            if not contents:
                raise ValueError('no content')
            return contents[0]['metadata']

        return self.__dict[key]

    def __iter__(self):
        return iter(self.__dict)

    def __len__(self):
        return len(self.__dict)

    def __contains__(self, item):
        return item in self.__dict

    def __str__(self):
        return str(self.__dict)

    def __eq__(self, o):
        return self.__dict.__eq__(o)

    def __ne__(self, o):
        return self.__dict.__ne__(o)

    def keys(self):
        return self.__dict.keys()

    def items(self):
        return self.__dict.items()

    def values(self):
        return self.__dict.values()

    def get(self, key, default=None):
        return self.__dict.get(key, default)

    ###############################
    # methods for file like object
    ###############################

    def read(self, size: int = -1) -> bytes:
        contents = self.__dict.get(KEY_CONTENTS, [])
        if not contents:
            raise ValueError('no body')
        return contents[0].read(size)  # type: ignore

    def close(self):
        contents = self.__dict.get(KEY_CONTENTS, [])
        for c in contents:
            c.close()


class HttpHeaders(Mapping):
    """HttpHeaders gives interface to get http headers.
    This class is readonly.

    HttpHeaders is partly compatible with with `werkzeug.datastructures.Headers`
    """

    def __init__(self, _headers: List[Dict] = []):
        """
        Input format of headers are,
         [
            {'key': 'accept', 'values': ['*/*']},
            {'key': 'user-agent', 'values': ['curl/7.64.1']},
            {'key': 'xyz', 'values': ['5', '3']},
            {'key': 'zzz', 'values': ['9']}
        ]
        """
        self.__list: List = []
        for item in _headers:
            key = item['key']
            values = item['values']
            for value in values:
                self.__list.append((key, value))

    def get_all(self, key: str):
        """Get all values for a given key.
        header names are case-insensitive.
        """
        ret = []
        lower_case_key = key.lower()
        for _key, _value in self.__list:
            if lower_case_key == _key.lower():
                ret.append(_value)
        return ret

    def __getitem__(self, key: str):
        """Returns the last item for the key out of the values.
        header names are case-insensitive.
        """
        values = self.get_all(key)
        if len(values) == 0:
            return None
        return values[-1]

    def __iter__(self):
        return iter(self.__list)

    def __len__(self):
        return len(self.__list)

    def __contains__(self, key):
        value = self.__getitem__(key)
        return value is not None

    has_key = __contains__

    def __eq__(self, o):
        def lowered(item):
            return (item[0].lower(),) + item[1:]
        return o.__class__ is self.__class__ and set(
            map(lowered, o.__list)
        ) == set(map(lowered, self.__list))

    def __ne__(self, o):
        return not self.__eq__(o)

    def keys(self):
        for item in self.__list:
            yield item[0]

    def items(self):
        for item in self.__list:
            yield item

    def values(self):
        for item in self.__list:
            yield item[1]

    def get(self, key, default=None):
        ret_val = self.__getitem__(key)
        if ret_val is not None:
            return ret_val
        return default



import asyncio
import functools
import os
import signal
import sys

from abejaruntime36.exception.exceptions import ModelSetupException
from abejaruntime36.logging.logger import get_logger
from abejaruntime36.model.model import setup_model
from abejaruntime36.server_protocol import SocketServerProtocol
from abejaruntime36 import version

DEFAULT_HANDLER = 'main:handler'

loop: asyncio.AbstractEventLoop
logger = get_logger()


def signal_handler(signame):
    logger.warning(f"runtime: signal [{signame}] received")
    global loop
    loop.stop()


def main():
    logger.info(f'start executing model with abeja-runtime-python (version: {version.VERSION})')

    socket_path = os.environ.get('ABEJA_IPC_PATH', None)
    if socket_path is None:
        logger.error("specify ABEJA_IPC_PATH for Inter Process Communication")
        sys.exit(1)

    handler = os.environ.get('HANDLER', DEFAULT_HANDLER)
    try:
        model = setup_model(handler)
    except ModelSetupException:
        logger.warning("failed to setup model", exc_info=True)
        sys.exit(1)

    global loop
    loop = asyncio.get_event_loop()
    for signame in ('SIGINT', 'SIGTERM'):
        loop.add_signal_handler(getattr(signal, signame),
                                functools.partial(signal_handler, signame))

    def get_protocol_factory() -> SocketServerProtocol:
        return SocketServerProtocol(model)

    server_coroutine = loop.create_unix_server(get_protocol_factory, socket_path)
    server = loop.run_until_complete(server_coroutine)

    try:
        loop.run_forever()
    except Exception:
        logger.exception("unexpected error")
    finally:
        server.close()
        loop.close()
        os.remove(socket_path)


if __name__ == "__main__":
    main()



import http
from abc import ABCMeta, abstractmethod
try:
    from collections import Iterable
except ImportError:
    from collections.abc import Iterable
from typing import List, Optional, BinaryIO

from abejaruntime36.entity.request import Request, Content
from abejaruntime36.entity.response import Response
from abejaruntime36.exception.exceptions import BadRequest, InternalServerError
from abejaruntime36.logging.logger import get_logger
from abejaruntime36 import version

logger = get_logger()

RESPONSE_KEY_STATUS_CODE = 'status_code'
RESPONSE_KEY_CONTENT_TYPE = 'content_type'
RESPONSE_KEY_CONTENT = 'content'
RESPONSE_KEY_METADATA = 'metadata'
DEFAULT_RESPONSE_STATUS_CODE = http.HTTPStatus.OK
DEFAULT_RESPONSE_CONTENT_TYPE = 'text/plain'
KEY_ABEJA_RUNTIME_VERSION = 'x-abeja-sys-meta-runtime-version'


class Converter(object, metaclass=ABCMeta):

    @abstractmethod
    def is_target(self, content_type: str) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def from_request(self, body: dict) -> Request:
        raise NotImplementedError()

    @abstractmethod
    def from_response(self, resp: dict) -> Response:
        raise NotImplementedError()

    def add_runtime_version(self, metadata: dict = None) -> dict:
        ret = metadata.copy() if metadata else {}
        ret[KEY_ABEJA_RUNTIME_VERSION] = version.VERSION
        return ret


class DefaultConverter(Converter):

    def is_target(self, content_type: str) -> bool:
        return True

    def from_request(self, req: dict) -> Request:
        http_method = req.get('method')
        content_type = req.get('content_type')
        headers = req.get('headers', [])
        ret = Request(str(http_method), str(content_type), headers, None)
        contents = req.get('contents', [])
        if contents is None:
            contents = []
        for c in contents:
            ret.append_content(self._from_content(c))
        return ret

    def from_response(self, resp: dict) -> Response:
        status_code = resp.get(RESPONSE_KEY_STATUS_CODE, DEFAULT_RESPONSE_STATUS_CODE)
        content_type = resp.get(RESPONSE_KEY_CONTENT_TYPE, DEFAULT_RESPONSE_CONTENT_TYPE)
        content = resp.get(RESPONSE_KEY_CONTENT)
        if content is not None:
            if not isinstance(content, Iterable) or isinstance(content, dict) or isinstance(content, str):
                content = [content]
        metadata = resp.get(RESPONSE_KEY_METADATA, None)
        metadata = super().add_runtime_version(metadata)
        return Response(content_type=content_type, contents=content, status_code=status_code, metadata=metadata)

    def _open_file(self, path: str) -> BinaryIO:
        body: BinaryIO
        try:
            body = open(path, "rb")
        except Exception as e:
            logger.error(f'failed to open file[{path}] that including request-body with error {e}')
            raise InternalServerError('failed to load request-body')

        return body

    def _from_content(self, content: dict) -> Content:
        content_type = content.get('content_type', None)
        file_name = content.get('file_name', None)
        form_name = content.get('form_name', None)
        metadata = content.get('metadata', None)
        path = content.get('path', None)
        body: Optional[BinaryIO] = None
        if path is not None:
            body = self._open_file(path)
        return Content(
            content_type=content_type, metadata=metadata, body=body, file_name=file_name, form_name=form_name
        )


converters: List[Converter] = [DefaultConverter()]


def convert_from_request(data: dict) -> Request:
    content_type = data.get('content_type', None)
    if content_type is None:
        raise BadRequest('content_type is null')

    for converter in converters:
        if converter.is_target(content_type):
            return converter.from_request(data)
    raise InternalServerError(f'unsupported content_type[{content_type}]')


def convert_from_response(data: dict) -> Response:
    if not isinstance(data, dict):
        raise TypeError('response of user-model is not dict')

    content_type = data.get('content_type', None)
    if content_type is None:
        return DefaultConverter().from_response(data)

    for converter in converters:
        if converter.is_target(content_type):
            return converter.from_response(data)
    raise InternalServerError(f'unsupported content_type[{content_type}]')



"""This module contains data types used by model developers."""
import array
import cgi
import csv
import http
import json
from collections import OrderedDict
from collections.abc import ByteString
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from abejaruntime36.helper.json_encoder import BetterJSONEncoder

OrderedKeyPairs = List[Tuple[str, Any]]

Content = Union[Dict, bytes]

Input = Union[OrderedKeyPairs, Content]

Metadata = Tuple[str, str]


class Response:
    """The data wrapper class to specify detail information of output content.

    Initialization
    ==============

    A ``Response`` object can be initialized with *content_type*, *contents* and *status_code*.
    The *contents* parameter is an *iterable* object containing elements of which
    type is one of below:

    - `bytes-like object`
    - ``dict``
    - ``list`` of key-value tuple (e.g. ``[(key, value), ...]``)

    >>> Response('text/plain', [b'Hello, World!'])
    <abejaruntime36.entity.response.Response object at 0x...>
    >>> Response('application/json', [{'apple': 103, 'orange': 250}])
    <abejaruntime36.entity.response.Response object at 0x...>
    >>> Response('application/json', [
    ...     [('apple', 103), ('orange', 250)],
    ...     [('apple', 345), ('orange', 87)]
    ... ])
    <abejaruntime36.entity.response.Response object at 0x...>

    You can specify status_code (e.g. HTTP Status Code) as the *status_code* parameter.
    A parameter can be ``http.HTTPStatus``:

    >>> Response('application/json', [{'status': 'bad request'}], status_code=HTTPStatus.BAD_REQUEST)
    <abejaruntime36.entity.response.Response object at 0x...>

    Default value of *status_code* is HTTPStatus.OK .

    You can specify HTTP headers as *metadata* parameter.
    A parameter can be ``dict``:

    >>> r = Response('text/plain; charset=utf-8', [b'Hello, World!'], metadata={'Retry-After', '100'})
    >>> r.metadata
    {'Retry-After', '100'}
    >>> r.mimetype
    'text/plain'

    Iterating contents
    ==================

    An instance of ``Response`` class is also an *iterator* of contents. You can read
    contents of a response by iterating partial contents with ``for`` loop.

    >>> r = Response([{'id': 1}, {'id': 2}])
    >>> for content in r:
    ...     print(content)
    ...
    {'id': 1}
    {'id': 2}

    Each partial content can be one of:

    - `bytes-like object`_
    - ``dict`` or ``collections.OrderedDict``

    If a partial content is ``list`` of key-value tuple, it will be converted to
    ``collections.OrderedDict``.

    >>> r = Response('application/json', [
    ...     [('apple', 103), ('orange', 250)],
    ...     [('apple', 345), ('orange', 87)]
    ... ])
    >>> list(r)
    [OrderedDict([('apple', 103), ('orange', 250)]), OrderedDict([('apple', 345), ('orange', 87)])]

    Like read-only I/O object, once you read the contents of a response,
    you can't read the contents anymore.

    >>> r = Response([{'id': 1}, {'id': 2}])
    >>> list(r)
    [{'id': 1}, {'id': 2}]
    >>> list(r)
    []

    .. _bytes-like object: https://docs.python.org/3/glossary.html#term-bytes-like-object
    """

    def __init__(self,
                 content_type: str,
                 contents: Optional[Iterable[Input]],
                 status_code: http.HTTPStatus = http.HTTPStatus.OK,
                 metadata: Dict[str, str] = None):
        if content_type is None or content_type.strip() == '':
            raise ValueError("need content_type")

        self.__content_type = content_type
        self.__status_code = status_code
        self.__iter = iter(contents) if contents else iter([])
        self.__metadata = metadata if metadata else {}

    def __iter__(self):
        return self

    def __next__(self) -> Content:
        element = next(self.__iter)

        # We have to convert an element to an instance of ``OrderedDict``
        # if it is a list of key-value tuples. However precisely checking
        # the type can be expensive, so we try to convert it if it is a list.
        if isinstance(element, list):
            try:
                return OrderedDict(element)
            except TypeError:
                pass
        return element  # type: ignore

    @property
    def content_type(self) -> str:
        return self.__content_type

    @property
    def status_code(self) -> http.HTTPStatus:
        return self.__status_code

    @property
    def metadata(self) -> Dict[str, str]:
        """Return an metadata.

        >>> r = Response("application/json", [], metadata={'Retry-After': '100'})
        >>> r.metadata
        {'Retry-After', '100'}
        """
        return self.__metadata

    @property
    def mimetype(self) -> Optional[str]:
        """Return the media type.

        This function parses *content_type* value and return MIME type part without options.
         e.g. ``application/json``.

        >>> r = Response("text/plain; charset=utf8", [])
        >>> r.mimetype
        'text/plain'
        """
        media_type, _options = cgi.parse_header(self.__content_type)
        return media_type.lower() if media_type else None

    class ByteBuffer(object):

        def __init__(self):
            self.buffer = []

        def extend(self, obj: bytes) -> None:
            self.buffer.append(obj)

        def write(self, content: str) -> None:
            self.buffer.append(content.encode('utf-8'))

        def empty(self) -> bool:
            """Return ``True`` if the queue is empty, ``False`` otherwise."""
            return len(self.buffer) == 0

        def tobytes(self) -> bytes:
            # Avoid copying buffered content if possible.
            if len(self.buffer) == 1:
                return self.buffer[0]
            else:
                return b''.join(self.buffer)

    def tobytes(self) -> bytes:
        r"""Convert the contents to ``bytes`` representation.

        Convertion Rules
        ================

        To convert its contents to ``bytes``, this method iterates elements and:

        1. If the element is a bytes-like object, no convertion happen.
        2. If ``Content-Type`` is ``text/csv``, use ``csv.DictWriter`` to
           convert it into csv format.
        3. Otherwise, convert it into json format by ``json.dumps``.
        4. To make it separated by newlines, insert newline characters appropriately.

        Examples
        ========

        *Bytes-like object*

        >>> r = Response('text/plain', [b'Hello, ', b'World!'])
        >>> r.tobytes()
        b'Hello, World!'

        *CSV*

        >>> r = Response('text/csv', [
        ...     [('apple', 103), ('orange', 250)],
        ...     [('apple', 345), ('orange', 87)]
        ... ])
        >>> r.tobytes()
        b'apple,orange\r\n103,250\r\n345,87\r\n'

        *JSON*

        >>> r = Response('application/json', [{'id': 123}])
        >>> r.tobytes()
        b'{"id": 123}'

        *JSON Lines*

        >>> r = Response('application/json', [
        ...     [('apple', 103), ('orange', 250)],
        ...     [('apple', 345), ('orange', 87)]
        ... ])
        >>> r.tobytes()
        b'{"apple": 103, "orange": 250}\n{"apple": 345, "orange": 87}'
        """
        buffer = Response.ByteBuffer()
        mimetype = self.mimetype
        csv_writer = None

        # TODO: Should we handle `str` object by encoding it with `charset` option in
        # `Content-Type`?
        for c in self:
            # Test an object is a bytes-like object.
            if isinstance(c, ByteString) or isinstance(c, array.array):
                buffer.extend(c)  # type: ignore
            elif isinstance(c, str):
                if not buffer.empty():
                    buffer.extend(b'\n')
                buffer.extend(c.encode('utf-8'))
            else:
                if mimetype == 'text/csv':
                    if not csv_writer:
                        csv_writer = csv.DictWriter(buffer, c.keys())
                        csv_writer.writeheader()

                    csv_writer.writerow(c)
                else:
                    if not buffer.empty():
                        # JSON serialization text should always be start at
                        # the begining of a line.
                        buffer.extend(b'\n')

                    buffer.write(json.dumps(c, cls=BetterJSONEncoder))

        return buffer.tobytes()



import json
import datetime
import numpy as np


class BetterJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime.datetime):
            return o.isoformat()
        elif isinstance(o, np.ndarray):
            return o.tolist()
        elif isinstance(o, np.integer):
            return int(o)
        elif isinstance(o, np.floating):
            return float(o)
        return super().default(o)



import asyncio
import binascii
import json
import os
import struct
import tempfile
from typing import Callable, Optional

from abejaruntime36.converter.converter import convert_from_request, convert_from_response
from abejaruntime36.entity.response import Response
from abejaruntime36.exception.exceptions import HttpError, InternalServerError
from abejaruntime36.logging.logger import get_logger

HEADER = b'\xAB\xE9\xA0'
VERSION = b'\x01'

logger = get_logger()

str_context = os.environ.get('CONTEXT', '{}')
CONTEXT = json.loads(str_context)


class SocketServerProtocol(asyncio.Protocol):
    transport: asyncio.Transport

    def __init__(self, model: Callable):
        self.model = model
        self.msg_part_header = bytearray()
        self.msg_part_len = bytearray()
        self.msg_part_body = bytearray()

    def connection_made(self, transport) -> None:
        self.transport = transport

    def _to_file(self, data: bytes) -> Optional[str]:
        if len(data) < 1:
            return None

        try:
            with tempfile.NamedTemporaryFile(delete=False) as fp:
                filename = fp.name
                fp.write(data)
        except IOError:
            logger.exception(f'failed to write response to {filename}')
            raise InternalServerError(f'failed to write response to {filename}')

        return filename

    def _send_response(self, res: dict) -> None:
        logger.debug(f'send response to proxy: {str(res)}')
        res_bytes = json.dumps(res).encode('utf-8')
        header = HEADER + VERSION + struct.pack('>I', len(res_bytes))
        self.transport.write(header)
        self.transport.write(res_bytes)

    def _transport_error(self, error: HttpError) -> None:
        res = {
            'content_type': 'application/json',
            'error_message': error.message,
            'status_code': error.status_code
        }
        return self._send_response(res)

    def _transport_response(self, resp: Response) -> None:
        try:
            filename = self._to_file(resp.tobytes())
        except HttpError as e:
            self._transport_error(e)
        except Exception:
            logger.exception("failed to transport:")
            return self._transport_error(InternalServerError("unexpected error occurred"))

        res = {
            'content_type': resp.content_type,
            'metadata': resp.metadata,
            'path': filename,
            'status_code': resp.status_code
        }
        self._send_response(res)

    def _parse_data(self, data: bytes) -> Optional[tuple]:
        data_len = len(data)
        header_len = len(self.msg_part_header)
        remain_header_len = 4 - header_len  # HEADER + VERSION = 4 byte
        if remain_header_len > 0:
            self.msg_part_header.extend(data[:remain_header_len])
        if data_len <= remain_header_len:
            return None

        len_len = len(self.msg_part_len)
        remain_len_len = 4 - len_len  # LENGTH-OF-BODY is 4 byte
        pos_len_end = remain_header_len + remain_len_len
        if remain_len_len > 0:
            self.msg_part_len.extend(data[remain_header_len:pos_len_end])
        if data_len <= pos_len_end:
            return None

        ret = None
        (length,) = struct.unpack('>I', self.msg_part_len)
        body_len = len(self.msg_part_body)
        remain_body_len = length - body_len
        pos_body_end = pos_len_end + length
        if remain_body_len > 0:
            self.msg_part_body.extend(data[pos_len_end:pos_body_end])
        if length == len(self.msg_part_body):
            ret = (self.msg_part_header.copy(), self.msg_part_body.decode('utf-8'))
            self.msg_part_header = bytearray()
            self.msg_part_len = bytearray()
            self.msg_part_body = bytearray()
        return ret

    def data_received(self, data: bytes) -> None:
        parsed_data = self._parse_data(data)
        if parsed_data is None:
            return

        header = parsed_data[0]
        if header != (HEADER + VERSION):
            hex_str = str(binascii.hexlify(header), 'utf-8')
            logger.error(f'request received with unsupported header or version: {hex_str}')
            return self._transport_error(InternalServerError('unexpected error occurred'))

        req_body = parsed_data[1]
        try:
            request = json.loads(req_body)
        except Exception:
            logger.exception(f'failed to decode request json from proxy: {req_body}')
            return self._transport_error(InternalServerError('unexpected error occurred'))

        logger.debug(f'request received from proxy: {str(request)}')
        try:
            req_contents = convert_from_request(request)
        except HttpError as e:
            self._transport_error(e)
            return

        try:
            resp = self.model(req_contents, CONTEXT)
            res_content = convert_from_response(resp)
        except Exception:
            logger.exception('unexpected error occurred in user model.')
            self._transport_error(InternalServerError('unexpected error occurred'))
            return
        finally:
            req_contents.close()

        self._transport_response(res_content)

    def connection_lost(self, exc) -> None:
        self.transport.close()



import http


class RuntimeException(Exception):
    def __init__(self, message):
        super().__init__()
        self.message = message


class ModelSetupException(RuntimeException):
    """Exception for failure while loading a model"""


class HttpError(RuntimeException):
    def __init__(self, status_code: http.HTTPStatus, message: str):
        super().__init__(message)
        self.status_code = status_code
        self.message = message


class BadRequest(HttpError):
    def __init__(self, message):
        super().__init__(http.HTTPStatus.BAD_REQUEST, message)


class InternalServerError(HttpError):
    def __init__(self, message):
        super().__init__(http.HTTPStatus.INTERNAL_SERVER_ERROR, message)



import logging
import sys


DEFAULT_USER_LOG_FORMAT = '%(levelname)s: %(message)s'


def __build_logger(service_name: str) -> logging.Logger:
    logger = logging.getLogger(service_name)
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(logging.Formatter(DEFAULT_USER_LOG_FORMAT))
    handler.setLevel(logging.DEBUG)
    logger.addHandler(handler)
    logger.propagate = False
    return logger


# The logger for runtime code
__LOGGER = __build_logger('runtime')


def get_logger() -> logging.Logger:
    """
    Return the logger instance for runtime.
    """
    return __LOGGER



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
import abejaruntime36.version


def _install_requires():
    requires = [
        'numpy',
    ]
    return requires


if __name__ == '__main__':
    description = 'ABEJA Platform Runtime for Python'
    setup(
        name='abejaruntime',
        version=abejaruntime36.version.VERSION,
        python_requires='>=3.6,<4.0',
        install_requires=_install_requires(),
        description=description,
        author='ABEJA Inc.',
        author_email='dev@abeja.asia',
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Intended Audience :: Developers",
            "Operating System :: OS Independent",
            "License :: OSI Approved :: Apache Software License",
        ],
        packages=find_packages(exclude=["tests.*", "tests"]),
        include_package_data=True,
        zip_safe=False,
        entry_points={
            "console_scripts": [
                "abeja-runtime-python = abejaruntime36.run:main"
            ]
        }
    )



import os
import subprocess
import sys
import importlib
from collections import deque
from pathlib import Path
from typing import Callable, Optional, List

from abejaruntime36.exception.exceptions import ModelSetupException
from abejaruntime36.logging.logger import get_logger

CONFIGURE_BASE_DIRECTORY = '.abeja'
WORKING_DIRECTORY_SEARCH_DEPTH = 3

logger = get_logger()


def _find_working_dir() -> Optional[str]:
    """
    Let the working directory be the path where the .abeja directory resides.

    """
    base_path = Path(os.curdir)
    base_depth = str(base_path.resolve()).count(os.sep)
    q: deque = deque([])
    q.append(base_path)
    while len(q) > 0:
        p = q.popleft()
        for child in p.iterdir():
            current_depth = str(child.resolve()).count(os.sep)
            if child.is_dir():
                if child.name == CONFIGURE_BASE_DIRECTORY:
                    return str(child.parent.resolve())
                else:
                    if (current_depth - base_depth) > WORKING_DIRECTORY_SEARCH_DEPTH:
                        continue
                    q.append(child)
    return None


def _install_packages() -> None:
    if os.path.exists('Pipfile'):
        logger.info('start installing packages from Pipfile')
        _install_from_pipfile()
        logger.info('packages are installed from Pipfile')
    elif os.path.exists('requirements.txt'):
        logger.info('start installing packages from requirements.txt')
        _install_from_requirements_txt()
        logger.info('packages are installed from requirements.txt')
    else:
        logger.info('requirements.txt/Pipfile not found, skip installing dependencies process')


def _install_from_pipfile() -> None:
    commands = ['pipenv', 'install', '--system']
    if not os.path.exists('Pipfile.lock'):
        commands.append('--skip-lock')
    _wait_subprocess(commands)


def _install_from_requirements_txt(requirements_path: str = 'requirements.txt') -> None:
    """installs python packages described in requirements
    Args:
        requirements_path:
    Returns:
        ``True`` if pip install succeeded.
        ``False`` if requirements.txt not found.
    Raises:
        CalledProcessError - failed to `pip install`
    """
    _wait_subprocess(['pip', 'install', '-r', requirements_path])


def _import_model(handler: str) -> Callable:
    """loads user model identified by handler
    Args:
        handler: path to the handler of user definied model
    Returns:
        function object of user defined model
    Raises:
        ImportError: failed to import user definied model
    """

    if handler.count(':') != 1:
        raise ImportError(
            'Possibility, your HANDLER[{}] parameter is wrong. '
            'handler needs one ":" separator. The format is <module>:<function>.'
            .format(handler))

    module_name, func_name = handler.split(':', 1)

    # Temporarily modify sys.path to include current working directory.
    if '' not in sys.path:
        sys.path.insert(0, '')

    try:
        model = getattr(importlib.import_module(module_name), func_name)
    except ImportError as e:
        raise ImportError(
            "Couldn't import a module named [{}] specified in HANDLER[{}]. "
            "Possibly, your HANDLER parameter or packaging (zip/tar) file structure is wrong. {}"
            .format(module_name, handler, e.__str__()))
    except AttributeError:
        raise ImportError(
            "Couldn't import a function named [{}] specified in HANDLER[{}]. "
            "Possibly, your HANDLER parameter or packaging (zip/tar) file structure is wrong."
            .format(func_name, handler))

    return model


def _wait_subprocess(command: List[str], max_wait: int = 5) -> None:
    """
    run and wait subprocess.
    Args:
        command (List[str]): command
        max_wait (int): max seconds to wait for subprocess to finish after closing stdout/err

    Returns:
        None

    Raises:
        RuntimeError
    """
    future = _run(command)
    for line in iter(future.stdout.readline, b''):
        logger.info(line.decode('utf-8'))

    try:
        returncode = future.wait(max_wait)
    except subprocess.TimeoutExpired:
        raise RuntimeError('subprocess did not exit after {} seconds'.format(max_wait))
    if returncode != 0:
        raise RuntimeError('subprocess finished with exit code {}'.format(future.returncode))


def _run(command: List[str]) -> subprocess.Popen:
    """
    run subprocess with specified *command*.
    Args:
        command (List[str]): command

    Returns:
        subprocess.Popen:
    """
    try:
        return subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    except (OSError, ValueError) as e:
        raise RuntimeError(e)


def setup_model(handler: str) -> Callable:
    working_dir = _find_working_dir()
    if working_dir:
        logger.debug('found .abeja directory in {}.'.format(working_dir))
        try:
            os.chdir(working_dir)
        except Exception as e:
            raise ModelSetupException(str(e))

    try:
        _install_packages()
    except RuntimeError as e:
        logger.exception(
            'error while installing from requirements.txt/Pipfile:')
        tb = sys.exc_info()[2]
        raise ModelSetupException(e).with_traceback(tb)

    try:
        logger.info("Loading user model handler with {}...".format(handler))
        model = _import_model(handler)
        return model
    except ImportError as e:
        logger.exception('error while importing user model')
        raise ModelSetupException('Exception occurred while loading handler function: {}'.format(str(e)))
